package POE::Component::Trickster::Streamer;
# $Id: Streamer.pm,v 1.4 2002/09/24 09:29:56 miyagawa Exp $

use strict;
use POE qw(Component::Server::TCP Filter::HTTPD Filter::Stream Component::Client::HTTP);
use FileHandle;
use File::Basename;
use File::Spec;
use HTTP::Response;
use HTTP::Request::Common;
use POSIX qw(SEEK_END);
use Time::HiRes qw(gettimeofday);
use Trickster::DiskJockey;
use Trickster::FileContainer;
use Trickster::MediaFile;
use Trickster::Listener;
use URI;
use URI::Escape;

use constant SERVER_NAME => 'server';

sub spawn {
    my($class, %args) = @_;
    POE::Session->create(
	package_states => [
	    $class => [ qw(_start scan_dir do_scan_dir poll enqueue dequeue skip stream
			   get_queue get_listeners download add_file
			   register unregister shutdown select_next
			   load_plugin register_plugin unregister_plugin dispatch_plugin
		       ) ],
	],
	heap => {
	    config => $args{Config},	# Trickster::Config
	    listeners => {},		# listerners ID
	    queue => [],		# files queue ([ Trickster::MediaFile ])
	    media => undef,		# current media file (Trickster::MediaFile)
	    handle => undef,		# current filehandle
	    schedule => undef,		# timestamp for scheduling
	    server => SERVER_NAME,	# server session name
	    shutdown => 0,		# flag to shutdown
	    container => Trickster::FileContainer->new,	# file container
	    dj => Trickster::DiskJockey->new($args{Config}),	# song selecter
	},
	args => [ \%args ],
    );
    my $player = delete $args{Alias};
    POE::Component::Trickster::Server->spawn(
	Config => $args{Config},
	Player => $player,
	Alias  => SERVER_NAME,
    );
    POE::Component::Client::HTTP->spawn(Alias => 'downloader');
}

sub _start {
    my($kernel, $heap, $args) = @_[KERNEL, HEAP, ARG0];
    $kernel->alias_set($args->{Alias});
    $kernel->yield('scan_dir');
}

sub scan_dir {
    my($kernel, $heap) = @_[KERNEL, HEAP];
    $heap->{container}->clear_all;
    $kernel->yield(do_scan_dir => $heap->{config}->get('files.root'));
    $kernel->yield('select_next');
    $kernel->yield('poll');
}

sub do_scan_dir {
    my($kernel, $heap, $dir) = @_[KERNEL, HEAP, ARG0];
    my $handle = DirHandle->new($dir);
    my $ext = $heap->{config}->get('files.ext');
    for my $f ($handle->read) {
	next if $f =~ /^\./;
	if (-d "$dir/$f") {
	    $kernel->yield(do_scan_dir => "$dir/$f");
	} elsif ($f =~ /$ext$/o) {
	    $heap->{container}->add_file(Trickster::MediaFile->new("$dir/$f"));
	}
    }
}

sub add_file {
    my($heap, $media) = @_[HEAP, ARG0];
    $heap->{container}->add_file($media);
}

sub poll {
    my($kernel, $heap) = @_[KERNEL, HEAP, ARG0];
    return if $heap->{shutdown};
    if (my $media = shift @{$heap->{queue}}) {
	if (@{$heap->{queue}} == 0) {
	    $kernel->yield('select_next');
	}
	if ($media->is_media_file) {
	    $heap->{media} = $media;
	    $heap->{handle} = FileHandle->new($media->file);
	    $kernel->yield(dispatch_plugin => got_start => [ $media ]);
	    $heap->{schedule} = gettimeofday;
	    return $kernel->yield('stream');
	} else {
	    $kernel->yield(dispatch_plugin => got_error => [
		$media->file . " is not a valid media file.",
	    ]);
	}
    }
    $kernel->yield('poll');
}

sub enqueue {
    my($kernel, $heap, $id) = @_[KERNEL, HEAP, ARG0];
    if ($id =~ m!^http://! and my $dir = $heap->{config}->get('files.incoming')) {
	$kernel->yield(download => $dir, $id);
	return;
    }
    my $media = $heap->{container}->get_file($id);
    if ($media) {
	push @{$heap->{queue}}, $media;
	$kernel->yield(dispatch_plugin => got_enqueue => [ $media ]);
    } else {
	$kernel->yield(dispatch_plugin => got_error => [ "$id not found" ]);
    }
}

sub dequeue {
    my($kernel, $heap, $file) = @_[KERNEL, HEAP, ARG0];
    my(@new_queue, $dequeued);
    for my $queue (@{$heap->{queue}}) {
	$queue->file eq $file ? $dequeued = $queue : push @new_queue, $queue;
    }
    $heap->{queue} = \@new_queue;
    $kernel->yield(dispatch_plugin => got_dequeue => [ $dequeued ]);
}

sub get_queue {
    my($kernel, $heap, $sender, $postback) = @_[KERNEL, HEAP, SENDER, ARG0];
    my $sub = ref($postback) ? $postback : sub {
	my($media, $queue) = @_;
	$kernel->post($sender->ID => $postback => $media, $queue);
    };
    $sub->($heap->{media}, $heap->{queue});
}

sub download {
    my($kernel, $heap, $dir, $url) = @_[KERNEL, HEAP, ARG0, ARG1];
    POE::Session->create(
	inline_states => {
	    _start => sub {
		$_[KERNEL]->post(downloader => request => got_response => GET $url);
	    },
	    got_response => sub {
		my($kernel, $heap, $req, $res) = @_[KERNEL, HEAP, ARG0, ARG1];
		my $file = File::Spec->catfile($dir, uri_unescape(basename(URI->new($url)->path)));
		my $handle = FileHandle->new("> $file");
		$handle->print($res->[0]->as_string);
		$handle->close;

		my $media = Trickster::MediaFile->new($file);
		$kernel->post(player => add_file => $media);
		$kernel->post(player => enqueue => $media->id);
	    },
	},
    );
}

sub skip {
    my($kernel, $heap) = @_[KERNEL, HEAP];
    seek $heap->{handle}, 0, SEEK_END if defined $heap->{handle};
    $kernel->yield(dispatch_plugin => 'got_skip');
}

sub stream {
    my($kernel, $heap) = @_[KERNEL, HEAP];
    my $bytes_read = $heap->{handle}->sysread(
	my $buffer = '', $heap->{config}->get('streaming.buffer_size'),
    );
    unless ($bytes_read) {
	# song ended, next song
	$kernel->yield(dispatch_plugin => got_done => [ $heap->{media} ]);
	delete $heap->{media};
	delete $heap->{handle};
	$kernel->yield('poll');
	return;
    }
    my $delay = $heap->{media}->delay_for($bytes_read);
    for my $listener (keys %{$heap->{listeners}}) {
	$kernel->post($listener => send_stream => $buffer);
    }
    $heap->{schedule} += $delay;
    $kernel->alarm_add(stream => $heap->{schedule});
}

sub register {
    my($kernel, $heap, $sender, $that_heap, $request) = @_[KERNEL, HEAP, SENDER, ARG0, ARG1];
    my $listener = Trickster::Listener->new(
	$sender->ID,
	$that_heap->{remote_addr} || $that_heap->{remote_ip},
	$request->header('User-Agent') || 'unknown',
    );
    $heap->{listeners}->{$sender->ID} = $listener;
    $kernel->yield(dispatch_plugin => got_register => [ $listener ]);
}

sub unregister {
    my($kernel, $heap, $sender, $why) = @_[KERNEL, HEAP, SENDER, ARG0];
    if ($heap->{listeners}->{$sender->ID}) {
	delete $heap->{listeners}->{$sender->ID};
	$kernel->yield(dispatch_plugin => got_unregiter => [ $sender->ID, $why ]);
    }
}

sub get_listeners {
    my($kernel, $heap, $sender, $postback) = @_[KERNEL, HEAP, SENDER, ARG0];
    my $sub = ref $postback ? $postback : sub {
	my @listeners = @_;
	$kernel->post($sender->ID, $postback => \@listeners);
    };
    $sub->(values %{$heap->{listeners}});
}

sub shutdown {
    my($kernel, $heap) = @_[KERNEL, HEAP];
    for my $listener (keys %{$heap->{listeners}}) {
	$kernel->post($listener => 'shutdown');
    }
    $kernel->post($heap->{server} => 'shutdown');
    $heap->{shutdown} = 1;
    seek $heap->{handle}, 0, SEEK_END if defined $heap->{handle};
}

sub select_next {
    my($kernel, $heap) = @_[KERNEL, HEAP];
    my $id = $heap->{dj}->select_next($heap->{container});
    $kernel->yield(enqueue => $id);
}

sub load_plugin {
    my($kernel, $heap, $plugin) = @_[KERNEL, HEAP, ARG0];
    eval { require "Trickster/Plugin/$plugin.pm"; };
    if ($@) {
	warn "plugin $plugin load fail; $@";
	return;
    }
    my $module = "Trickster::Plugin::$plugin";
    my $object = $module->new($heap->{config});
    $object->init;
}

sub register_plugin {
    my($kernel, $heap, $sender) = @_[KERNEL, HEAP, SENDER];
    $heap->{plugins}->{$sender->ID} = 1;
    $kernel->refcount_increment($sender->ID, __PACKAGE__);
}

sub unregister_plugin {
    my($kernel, $heap, $sender) = @_[KERNEL, HEAP, SENDER];
    delete $heap->{plugins}->{$sender->ID};
    $kernel->refcount_decrement($sender->ID, __PACKAGE__);
}

sub dispatch_plugin {
    my($kernel, $heap, $action, $args) = @_[KERNEL, HEAP, ARG0, ARG1];
    $args ||= [];
    for my $plugin (keys %{$heap->{plugins}}) {
	$kernel->post($plugin => $action => @$args);
    }
}

#--------------------------------------------------
# for HTTP streaming server
package POE::Component::Trickster::Server;
use POE::Session;

sub spawn {
    my($class, %args) = @_;
    POE::Component::Server::TCP->new(
	Alias => $args{Alias},
	Port  => $args{Config}->get('streaming.port'),
	ClientFilter => 'POE::Filter::HTTPD',
	ClientDisconnected => sub {
	    $_[KERNEL]->post($args{Player} => unregister => 'client disconnected');
	},
	ClientError => sub {
	    _shutdown_client($_[KERNEL], \%args, 'client connection closed');
	},
	ClientInput => sub {
	    my($kernel, $heap, $request) = @_[KERNEL, HEAP, ARG0];
	    if ($request->isa('HTTP::Response')) {
		# error!
		$heap->{client}->put($request);
		_shutdown_client($kernel, \%args, 'invalid HTTP request');
		return;
	    }
	    my $response = HTTP::Response->new(200);
	    $response->push_header('Content-Type' => 'audio/x-mpeg');
	    $response->push_header('X-Audiocast-Name' => $args{Config}->get('streaming.credit'));
	    $heap->{client}->put($response);

	    # send IP addy and User-Agent:
	    $kernel->post($args{Player} => register => $heap, $request);
	    $kernel->yield('start_stream');
	},
	InlineStates => {
	    start_stream => sub {
		my($kernel, $heap) = @_[KERNEL, HEAP];
		$heap->{client}->set_output_filter(POE::Filter::Stream->new);
	    },
	    send_stream  => sub {
		my($kernel, $heap, $buffer) = @_[KERNEL, HEAP, ARG0];
		unless ($heap->{client}) {
		    # client disconnected for some reason
		    _shutdown_client($kernel, \%args, 'unknown reason');
		    return;
		}
		my $pending_bytes = $heap->{client}->get_driver_out_octets;
		if ($pending_bytes >= $args{Config}->get('streaming.buffer_size') * 8) {
		    # client is too slow
		    _shutdown_client($kernel, \%args, 'client too slow connection');
		    return;
		}
		$heap->{client}->put($buffer);
	    },
	},
    );
}

sub _shutdown_client {
    my($kernel, $args, $why) = @_;
    $kernel->post($args->{Player} => unregister => $why);
    $kernel->yield('shutdown');
}

1;
__END__

=head1 NAME

POE::Component::Trickster::Streamer - Trickster Streaming Server

=head1 SYNOPSIS

  use POE qw(Component::Trickster::Streamer);
  POE::Component::Trickster::Streamer->spawn(port => $port, buffer => $buffer);

=head1 DESCRIPTION

Trickster::Streamer is a POE component for Trickster streaming
server.

=head1 AUTHOR

Tatsuhiko Miyagawa E<lt>miyagawa@bulknews.netE<gt>

This library is free software; you can redistribute it and/or modify
it under the same terms as Perl itself.

=head1 SEE ALSO
v
L<POE>, L<POE::Component::Trickster::Controller>

=cut
